/*
   Class Name:

      CQuaternion.

   Created by:

      Allen Sherrod (Programming Ace of www.UltimateGameProgramming.com).

   Description:

      This class is used to create a quaternion to be used with rotations.
*/


#ifndef CQUATERNION_H
#define CQUATERNION_H

#define PI 3.141592654                                      // Define for PI.
#define GET_RADIANS(degree) (float)(degree / 180.0f * PI)   // Will convert degrees to radians.

#include<math.h>                                            // Math header file.
#include"CVector.h"


class CQuaternion 
{
   public:
      CQuaternion()
         {
            // Initialize each member variables.
            x = y = z = 0.0f;
            w = 1.0f;
         }
	   
	   CQuaternion(float xAxis, float yAxis, float zAxis, float wAxis) 
	      {
            // Initialize each member variables.
		      x = xAxis; y = yAxis; z = zAxis;
            w = wAxis;
	      }

      void operator=(const CQuaternion &q)
         {
            // This will make this quaternion equal to q.
            w = q.w; x = q.x; y = q.y; z = q.z;
         }

      CQuaternion operator*(const CQuaternion &q)
         {
            return CQuaternion(w * q.x + x * q.w + y * q.z - z * q.y,
                               w * q.y - x * q.z + y * q.w + z * q.x,
                               w * q.z + x * q.y - y * q.x + z * q.w,
                               w * q.w - x * q.x - y * q.y - z * q.z);
         }

      CQuaternion operator+(const CQuaternion &q)
         {
            CQuaternion result;

            CVector4 v1(x, y, z), v2(q.x, q.y, q.z), cross, v3;

            cross.CrossProduct(v2, v1);
            v1 *= q.w;
            v2 *= w;
            
            v3 = v1 + v2;
            v3 = cross + v3;

            v3.w = w * q.w - (x * q.x + y * q.y + z * q.z);
            
            result.x = v3.x; result.y = v3.y; result.z = v3.z; result.w = v3.w;

            return result;
         }

      void CreateQuatFromAxis(CVector4 a, float phi)
         {
            float sine = 0.0f;
            
            phi = phi * 0.5f;
            sine = (float)sin(phi);
            
            a.Normal();

            x = a.x; y = a.y; z = a.z;

            x *= sine;
            y *= sine;
            z *= sine;
            w = (float)cos(phi);
         }

      float Length()
         {
            return (float)sqrt(x * x + y * y + z * z + w * w);
         }


      void Normal()
         {
            float len = Length();
            len = 1 / len;

            x = x * len;
            y = y * len;
            z = z * len;
            w = w * len;
         }


      CQuaternion Conjugate()
         {
            return CQuaternion(-x, -y, -z, w);
         }


      CQuaternion CrossProduct(const CQuaternion &q)
         {
            CQuaternion crossProduct;

            crossProduct.x = w * q.x + x * q.w + y * q.z - z * q.y;
            crossProduct.y = w * q.y + x * q.z + y * q.w - z * q.x;
            crossProduct.z = w * q.z + x * q.y + y * q.x - z * q.w;
            crossProduct.w = w * q.w - x * q.x - y * q.y - z * q.z;

            return crossProduct;
         }

	   void Rotatef(float amount, float xAxis, float yAxis, float zAxis)
         {
            // Pretty much what is going on here is that if there is any axis that is not
            // a 0 or a 1 (meaning its not normalized) then we want to normalize it.
            // I created a if statement because I thought this would be better than normalizing
            // it every time this function is called, which would result in a lot of expensive
            // sqrt() call.  This is just in case the user forgot to use only normalized values.
            if((xAxis + yAxis + zAxis) != 1)
            {
               float length = (float)sqrt(xAxis * xAxis + yAxis * yAxis + zAxis * zAxis);
               xAxis /= length; yAxis /= length; zAxis /= length;
            }

            // Convert the angle degrees into radians.
            float angle = GET_RADIANS(amount);

            // Call this once for optimization.
	         float sine = (float)sin(angle / 2.0f);

            // Create the quaternion.
	         x = xAxis * sine;
	         y = yAxis * sine;
	         z = zAxis * sine;
            w = (float)cos(angle / 2.0f);
         }

	   void CreateMatrix(float *pMatrix)
         {
            // Error checking...
	         if(!pMatrix)
               return;

	         // Calculate the first row.
            pMatrix[0]  = 1.0f - 2.0f * (y * y + z * z); 
            pMatrix[1]  = 2.0f * (x * y + z * w);
            pMatrix[2]  = 2.0f * (x * z - y * w);
            pMatrix[3]  = 0.0f;  

            pMatrix[4]  = 2.0f * (x * y - z * w);  
            pMatrix[5]  = 1.0f - 2.0f * (x * x + z * z); 
            pMatrix[6]  = 2.0f * (z * y + x * w);  
            pMatrix[7]  = 0.0f;  

            pMatrix[8]  = 2.0f * (x * z + y * w);
            pMatrix[9]  = 2.0f * (y * z - x * w);
            pMatrix[10] = 1.0f - 2.0f * (x * x + y * y);  
            pMatrix[11] = 0.0f;  

            pMatrix[12] = 0;  
            pMatrix[13] = 0;  
            pMatrix[14] = 0;  
            pMatrix[15] = 1.0f;
         }

      void Slerp(const CQuaternion &q1, const CQuaternion &q2, float t)
      {
         float cosTheta = 0.0f;
         float sinTheta = 0.0f;
         float beta = 0.0f;
         float q2Array[4];

         // Temporary array to hold second quaternion.
         q2Array[0] = q2.x; q2Array[1] = q2.y; q2Array[2] = q2.z; q2Array[3] = q2.w;

         cosTheta = q1.x * q2.x + q1.y * q2.y + q1.z * q2.z + q1.w * q2.w;

         if(cosTheta < 0.0f)
            {
               // Flip sigh if so.
               q2Array[0] = -q2Array[0]; q2Array[1] = -q2Array[1];
               q2Array[2] = -q2Array[2]; q2Array[3] = -q2Array[3];
               cosTheta = -cosTheta;
            }

         beta = 1.0f - t;

         if(1.0f - cosTheta > 0.001f)
            {
               // We are using spherical interpolation.
               cosTheta = (float)acos(cosTheta);
               sinTheta = 1.0f / (float)sin(cosTheta);
               beta = (float)sin(cosTheta * beta) * sinTheta;
               t = (float)sin(cosTheta * t) * sinTheta;
            }

         // Interpolation.
         x = beta * q1.x + t * q2Array[0];
         y = beta * q1.y + t * q2Array[1];
         z = beta * q1.z + t * q2Array[2];
         w = beta * q1.w + t * q2Array[3];
      }

	   float x, y, z, w;                      // x, y , z, and w axis.
};

#endif


// Copyright March 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com