/*
   Simple Keyframe Animation Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to perform simple key frame animation on a simple object.
   The animation will be simple rotation and translate.  We will have an array of rotation
   and translation values that we will use linear interpolation on to get the in between
   values.
*/


// Demo: Simple Keyframe Animation.


#define WIN32_LEAN_AND_MEAN
#define VC_LEANMEAN

#include<d3dx9.h>
#include<mmsystem.h>
#include"CVector.h"
#include"CQuaternion.h"
#include"CMatrix.h"


// Function Prototypes...
bool InitializeDirect3D(HWND hwnd, bool fullscreen);// Initialize Direct3D.
bool InitializeObject();                        // Initialize object we are displaying.
void RenderScene();                             // Render the 3D scene.
void SetPerpective();                           // Set the screen perpective.
void ShutdownDirect3D();                        // Shutdown Direct3D.


LPDIRECT3D9 Direct3D_Object = NULL;             // Direct3D object.
LPDIRECT3DDEVICE9 D3D_Device = NULL;            // Device object.

LPD3DXBUFFER MaterialBuffer = NULL;             // Material buffer.
LPD3DXMESH Teapot = NULL;                       // Mesh object.
D3DMATERIAL9 TeapotMaterial;                    // Objects material.

D3DLIGHT9 light;                                // Main light object.

// Used for time based movements.
DWORD StartTime = 0;

// Structure used to hold each frames rotation and translation values for our object.
typedef struct KeyFrame
{
   DWORD time;                            // Start time of this keyframe.
   CVector4 rot;                          // Rotation amount of this frame.
   CVector4 trans;                        // Translation amount of this frame.
} KeyFrame;

// List of keyframes in out animation.
KeyFrame frames[4];

// Globals for the width and height.
int gWidth = 0, gHeight = 0;


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   int width, height;                     // The window width and height.

   switch(message)
      {
		   case WM_KEYUP:
            // If the user presses the escape key then exit the application.
            if(wParam == VK_ESCAPE)
               PostQuitMessage(0);
            break;


         case WM_DESTROY:
         case WM_CLOSE:
            PostQuitMessage(0);
            break;

         case WM_SIZE:                       // re-size message.
            height = HIWORD(lParam);         // This gets the height of the window.
            width = LOWORD(lParam);          // This gets the width of the window.

            if(height==0)                    // we don't want it to be possible for a
               {                             // height of 0.  If it is 0 me make it 1.
                  height = 1;
               }

            gWidth = width;
            gHeight = height;

            return 0;
            break;

         default:
            break;
       }

    return DefWindowProc( hwnd, message, wParam, lParam );
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                                           // Message object.
   HWND hwnd;                                         // Handle to the window.
   WNDCLASSEX windowClass;                            // Window class object.
   bool done = false;                                 // True then exit.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);           // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;       // style of the window.
   windowClass.lpfnWndProc = WndProc;                 // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                        // Extra class information.
   windowClass.cbWndExtra = 0;                        // Extra window information.
   windowClass.hInstance = hInstance;                 // Handle of application Instance.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);// Window Icon.
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW); // Load mouse cursor.
   windowClass.hbrBackground = NULL;                  // Background color.
   windowClass.lpszMenuName = NULL;                   // Menu.
   windowClass.lpszClassName = "UGPClass";            // Name of the window class.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);// Minimized window icon.

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // After your class has been registered then you are ready to create your window.
   hwnd = CreateWindowEx(NULL,                                       // The extended style.
                         "UGPClass",                                 // Window class.
                         "Simple Keyframe Demo - by The Programming Ace",// Window name.
                          WS_OVERLAPPEDWINDOW | WS_VISIBLE |         // Window style.
                          WS_SYSMENU |WS_CLIPCHILDREN |              // Window style.
                          WS_CLIPSIBLINGS,                           // Window style.
                          100, 100,                                  // X, Y coords.
                          640, 480,                                  // Window size.
                          NULL,                                      // Handle to parent window.
                          NULL,                                      // Menu.
                          hInstance,                                 // Handle to app instance.
                          NULL);                                     // Pointer to window.

   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW);    // Show the window.
   UpdateWindow(hwnd);           // Update its display.

   done = false;                 // false = run program, true means stop.

   // Initialize Direct3D.  If fails then we don't want to run the application.
   if(!InitializeDirect3D(hwnd, false)) done = true;
   
   // Application loop.
   while(!done)
      {
         if(PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     done = true;
                  }

               TranslateMessage(&msg);
               DispatchMessage(&msg);
            }
         else
            {
               RenderScene();
            }
      }

    // Here we shutdown Direct 3D.
	ShutdownDirect3D();

   // Here we unregister the window class with the OS.
   UnregisterClass("UGPClass", windowClass.hInstance);

   return (int)msg.wParam;
}


bool InitializeDirect3D(HWND hwnd, bool fullscreen)
{
   // This object will allow us to set the display mode of the screen.
   D3DDISPLAYMODE DisplayMode;

   // This will allow us to set the parameters of the screen.
   D3DPRESENT_PARAMETERS Present_Parameters;

   // This is used to get the capabilities of the hardware.
   D3DCAPS9 D3DCaps;

   // It is always a good idea to clear out memory in object although not necessary.
   ZeroMemory(&Present_Parameters, sizeof(Present_Parameters));

   // Create the Direct3D object to get everything started.
   Direct3D_Object = Direct3DCreate9(D3D_SDK_VERSION);

   // Error checking.  Make sure that it was successful.
   if(Direct3D_Object == NULL)
      {
         MessageBox(NULL, "Error, couldn't initialize DirectX!?!",
                    "Error!", MB_OK);
         return false;
      }

   // This function will get the display mode of the device and place it in DisplayMode.
   if(FAILED(Direct3D_Object->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &DisplayMode)))
      {
         MessageBox(NULL, "Error setting the display mode.", "Error!", MB_OK);
         return false;
      }

   // Get the capabilities of the hardware.
	if(FAILED(Direct3D_Object->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &D3DCaps)))
		return false;

   // Test which is supported, hardware or software vertex processing.
	DWORD VertexProcessing = 0;

	if(D3DCaps.VertexProcessingCaps != 0)
		VertexProcessing |= D3DCREATE_HARDWARE_VERTEXPROCESSING;
	else
		VertexProcessing |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;

   // Here we are setting the applications parameters...
   if(fullscreen)
      {
         Present_Parameters.Windowed = FALSE;               // Window mode (fullscreen).
         Present_Parameters.BackBufferWidth = DisplayMode.Width;
         Present_Parameters.BackBufferHeight = DisplayMode.Height;
      }
   else
      Present_Parameters.Windowed = TRUE;                   // Window mode (not fullscreen).
   Present_Parameters.SwapEffect = D3DSWAPEFFECT_DISCARD;   // Dealing with animation (see doc).
   Present_Parameters.BackBufferFormat = DisplayMode.Format;// Render to the area of the screen.
   Present_Parameters.BackBufferCount = 1;                  // Number of back buffers.
   Present_Parameters.EnableAutoDepthStencil = TRUE;        // Check documentation.
   Present_Parameters.AutoDepthStencilFormat = D3DFMT_D16;  // Check documentation.

   // Now we must create the rendering device.
   if(FAILED(Direct3D_Object->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hwnd,
                        VertexProcessing, &Present_Parameters, &D3D_Device)))
      {
         MessageBox(NULL, "CreateDevice() failed!  Make sure you have DirectX 9.",
                    "Error!", MB_OK);
         return false;
      }

   // One last check to be sure.
   if(D3D_Device == NULL)
      {
         MessageBox(NULL, "D3D_Device is equal to NULL!?!", "Error!", MB_OK);
         return false;
      }

   // Set D3D states.
   D3D_Device->SetRenderState(D3DRS_LIGHTING, TRUE);        // Lighting on.
	D3D_Device->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);// Culling off.
   D3D_Device->SetRenderState(D3DRS_ZENABLE, TRUE);         // Enable depth testing.

   // Here we set up the object this demo will render (for example a polygon).
   if(!InitializeObject()) return false;

   // Set the perpective.
   SetPerpective();

   return true;
}


bool InitializeObject()
{
   // Setup the light source.
   light.Type = D3DLIGHT_DIRECTIONAL;
   light.Direction = D3DXVECTOR3(0.0f, 0.5f, 1.0f);

   light.Diffuse.r = light.Diffuse.g = light.Diffuse.b = light.Diffuse.a = 1.0f;
   light.Specular.r = light.Specular.g = light.Specular.b = light.Specular.a = 1.0f;

   D3D_Device->SetLight(0, &light);
   D3D_Device->LightEnable(0, TRUE);

   // Setup the material properties for the teapot.
   ZeroMemory(&TeapotMaterial, sizeof(D3DMATERIAL9));
   TeapotMaterial.Diffuse.r = TeapotMaterial.Ambient.r = 0.6f;
   TeapotMaterial.Diffuse.g = TeapotMaterial.Ambient.g = 0.6f;
   TeapotMaterial.Diffuse.b = TeapotMaterial.Ambient.b = 0.7f;
   TeapotMaterial.Specular.r = 0.4f;
   TeapotMaterial.Specular.g = 0.4f;
   TeapotMaterial.Specular.b = 0.4f;
   TeapotMaterial.Power = 8.0f;

   // To create the teapot we simple call D3DXCreateTeapot and the rest is done for us.
   if(FAILED(D3DXCreateTeapot(D3D_Device, &Teapot, &MaterialBuffer)))
      return false;

   // Set the first frame's start time, rotation value, and translation position.
   frames[0].time = 0;
   frames[0].rot = CVector4(0.0f, 1, 0, 0);
   frames[0].trans = CVector4(0.0f, 0.0f, 0.0f);

   // Set the second frame's start time, rotation value, and translation position.
   frames[1].time = 800;
   frames[1].rot = CVector4(130.0f, 1, 0, 0);
   frames[1].trans = CVector4(5.0f, 5.0f, 0.0f);

   // Set the third frame's start time, rotation value, and translation position.
   frames[2].time = 1600;
   frames[2].rot = CVector4(260.0f, 1, 0, 0);
   frames[2].trans = CVector4(-5.0f, 5.0f, 0.0f);

   // Set the fourth frame's start time, rotation value, and translation position.
   frames[3].time = 2400;
   frames[3].rot = CVector4(360.0f, 1, 0, 0);
   frames[3].trans = CVector4(0.0f, 0.0f, 0.0f);

   // Initialize the start time of our simulation.
   StartTime = timeGetTime();

   return true;
}


void SetPerpective()
{
   // This matrix will do nothing more but tell Direct3D how far back it can see
   // and how close it can see.
   D3DXMATRIXA16 Projection;

   // Set the perspective information.
   D3DXMatrixPerspectiveFovLH(&Projection, 45.0f, (float)(gWidth/gHeight), 0.1f, 500.0f);

   // Set the projection.
   D3D_Device->SetTransform(D3DTS_PROJECTION, &Projection);
}


void SetRotation()
{
   // Get the current time into the animation.
   DWORD Time = timeGetTime() - StartTime;

   // Bounds the time to the animation time of the last frame.
   Time %= frames[3].time;

   // Determine which key frame to use.
   DWORD KeyF1 = 0, KeyF2 = 0, TimeDiff = 0;

   // We loop through the list of key frames to find which one we are in.
   // This is determined by whether this time is greater than or equal to
   // a frame but less than the next.
   for(DWORD i = 0; i < 4; i++)
      {
          if(Time >= frames[i].time) KeyF1 = i;
      }

   // The second key frame is basically the one after the frame we are in.
   // If we are in the last frame, then the next frame is the first one.
   KeyF2 = (KeyF1 == 3) ? KeyF1 : KeyF1 + 1;

   // Difference in the time between the two frames.
   TimeDiff = frames[KeyF2].time - frames[KeyF1].time;
   
   // Error checking, keep us from dividing by 0 or less than 0.
   if(!TimeDiff) TimeDiff = 1;

   // Here we will calculate the scalar that will be used to interpolate between 2 values.
   float Scalar = ((float)(Time - frames[KeyF1].time) / (float)TimeDiff);

   // quaternionRot will store the interpolation rotation value as a quaternion.  interpolVec
   // will store the interpolation values temporarily.  matRot will be used to turn the
   // quaternion rotation into a matrix, matTrans will store the interpolated translation
   // values, and matFinal will store the combination between the rotation and translate matrix.
   CQuaternion quaternionRot;
   CVector4 interpolVec;
   CMatrix4x4 matRot, matTrans, matFinal;

   // Use interpolation to get the interpolated rotation value.
   interpolVec.x = frames[KeyF1].rot.x + (frames[KeyF2].rot.x - frames[KeyF1].rot.x) * Scalar;
   interpolVec.y = frames[KeyF1].rot.y + (frames[KeyF2].rot.y - frames[KeyF1].rot.y) * Scalar;
   interpolVec.z = frames[KeyF1].rot.z + (frames[KeyF2].rot.z - frames[KeyF1].rot.z) * Scalar;
   interpolVec.w = frames[KeyF1].rot.w + (frames[KeyF2].rot.w - frames[KeyF1].rot.w) * Scalar;
   
   // Turn the interpolated rotation into a quaternion then a matrix.
   quaternionRot.Rotatef(interpolVec.x, interpolVec.y, interpolVec.z, interpolVec.w);
   quaternionRot.CreateMatrix(matRot.matrix);

   // Interpolation the translation values.
   interpolVec.x = frames[KeyF1].trans.x + (frames[KeyF2].trans.x - frames[KeyF1].trans.x) * Scalar;
   interpolVec.y = frames[KeyF1].trans.y + (frames[KeyF2].trans.y - frames[KeyF1].trans.y) * Scalar;
   interpolVec.z = frames[KeyF1].trans.z + (frames[KeyF2].trans.z - frames[KeyF1].trans.z) * Scalar;

   // Create the matrix.
   matTrans.Translate(interpolVec.x, interpolVec.y, interpolVec.z);

   // Combine the two matrices.
   matFinal = matRot * matTrans;

   // Set the matrix.
   D3D_Device->SetTransform(D3DTS_WORLD, (D3DMATRIX*)matFinal.matrix);
}


void SetView()
{
   // This will hold the view matrix.
   D3DXMATRIXA16 View;

   D3DXVECTOR3 Eye(0.0f, 0.0f, -10.0f);    // Position of view.
   D3DXVECTOR3 LookAt(0.0f, 0.0f, 0.0f);   // Point view is looking at.
   D3DXVECTOR3 Up(0.0f, 1.0f, 0.0f);       // Which way is up.

   // Set all the values this matrix will need to know (position, look at point, etc).
   D3DXMatrixLookAtLH(&View, &Eye, &LookAt, &Up);

   // Move the "view" to the position and values we set above.
   D3D_Device->SetTransform(D3DTS_VIEW, &View);
}


void RenderScene()
{
   // Clear the screen to black.
   D3D_Device->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                     D3DCOLOR_XRGB(0,0,0), 1.0f, 0);
    
   // Tell Direct 3D we are now going to start drawing.
   D3D_Device->BeginScene();

	   // Set the view matrix (view position).
      SetView();

      // Rotate scene
      SetRotation();

      // Set the material and draw the Teapot.
      D3D_Device->SetMaterial(&TeapotMaterial);
      Teapot->DrawSubset(0);
    
   // We are done drawing for this scene.
   D3D_Device->EndScene();
    
   // Swap the old frame with the new one.
   D3D_Device->Present(NULL, NULL, NULL, NULL);
}


void ShutdownDirect3D()
{
   // Here we release the Direct3D objects.
    if(D3D_Device != NULL)
      {
         D3D_Device->Release();
         D3D_Device = NULL;
      }

   if(Direct3D_Object != NULL)
      {
         Direct3D_Object->Release();
         Direct3D_Object = NULL;
      }

   if(MaterialBuffer != NULL)
      {
         MaterialBuffer->Release();
         MaterialBuffer = NULL;
      }

   if(Teapot != NULL)
      {
         Teapot->Release();
         Teapot = NULL;
      }
}


// Recap:

// This demo is not that advance.  All we are doing is storing a list of rotation and
// translation values in an array (which we could have loaded from a file) then using
// linear interpolation to get the objects position over time.


// Copyright September 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com